#!/bin/bash

# mktorfw.sh - create an linux iptables firewall for tor client

#
# Copyright (c) 2007, Gregory Fleischer (gfleischer@gmail.com)
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# 3. Neither the name of the author nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

set -e -u

show_help()
{
    echo ""
    echo "Usage: $0 [options] -i <iface>"
    echo "  -d                 debug"
    echo "  -u                 update tor rules"
    echo "  -i  <iface>        the external interface"
    echo "  -t                 comma-separated list of trusted interfaces"
    echo "  -l                 comma-separated list of locally trusted ports"
    echo "  -h                 show usage"
    echo ""

}

dbg=0
update=0
iface=""
trusted=""
localports=""
while getopts "dui:t:l:h" opt; do
    case $opt in 
	d)
	    dbg=1
	    ;;
	u)
	    update=1
	    ;;
	i)
	    iface=$OPTARG
	    ;;
	t)  
	    trusted=$OPTARG
	    ;;
	l)  
	    localports=$OPTARG
	    ;;
	h)
	    show_help
	    exit 0
	    ;;
	\?)
	    echo "invalid option" >&2
	    show_help
	    exit 1
	    ;;
    esac
done

shift $(($OPTIND - 1))

if [ -z "$iface" ]; then
    echo "missing interface" >&2
    show_help
    exit 1
fi

if [ ! -e "/proc/sys/net/ipv4/conf/$iface" ]; then
    echo "interface [$iface] doesn't seem to exist" >&2
    exit 1
fi

IPTABLES="/sbin/iptables"
if [ 1 = $dbg ]; then
    IPTABLES="$IPTABLES -v"
fi

EXT_IFACE="$iface"

# source util lib
. libutiltor.sh

set_default_policy()
{
    $IPTABLES -P INPUT DROP
    $IPTABLES -P OUTPUT DROP
    $IPTABLES -P FORWARD DROP
}

flush_all_rules()
{
    for i in filter; do
	$IPTABLES -t $i -F
	$IPTABLES -t $i -X
    done
}

create_reject_drop_rules()
{
    # reject
    $IPTABLES -N reject_and_log
    $IPTABLES -A reject_and_log -j LOG --log-prefix "rejecting packet: "
    $IPTABLES -A reject_and_log -p tcp -j REJECT --reject-with tcp-reset
    $IPTABLES -A reject_and_log -j REJECT --reject-with icmp-port-unreachable

    # drop
    $IPTABLES -N drop_and_log
    $IPTABLES -A drop_and_log -j LOG --log-prefix "dropping packet: "
    $IPTABLES -A drop_and_log -j DROP
}

create_local_port_rules()
{

    # local ports
    $IPTABLES -N local_ports

    if [ '*' = "$privoxyhost" ]; then
	$IPTABLES -A local_ports -p tcp --dport $privoxyport -j ACCEPT
    else
	$IPTABLES -A local_ports -p tcp -d $privoxyhost --dport $privoxyport -j ACCEPT
    fi

    $IPTABLES -A local_ports -p tcp -d $sockshost --dport $socksport -j ACCEPT

    if [ 0 != $controlport ]; then
	$IPTABLES -A local_ports -p tcp -d localhost --dport $controlport -j ACCEPT
    fi

    if [ -n "$localports" ]; then
	for port in `echo "$localports" | sed -e 's/,/\n/g'`; do
	    $IPTABLES -A local_ports -p tcp -d localhost --dport $port -j ACCEPT
	done
    fi
}

add_tor_rules()
{

    method="${1}"

    # create cache directory
    if [ ! -d /var/cache/torfw ]; then
	install -m0755 -vd /var/cache/torfw
    fi

    # read from cached routers
    cat /var/lib/tor/cached-routers* | grep "^router " | \
	awk '(!$4~/0/) {print "$IPTABLES -A is_tor -p tcp -d "$3" --dport "$4" -j ACCEPT"} \
        (!$6~/0/) {print "$IPTABLES -A is_tor -p tcp -d "$3" --dport "$6" -j ACCEPT"}' \
	| sort -u > /var/cache/torfw/tor-rules.new

    # use the diff to determine actions to apply
    if [ 'U' = "$method" ] && [ -f /var/cache/torfw/tor-rules ]; then

	(diff -N /var/cache/torfw/tor-rules /var/cache/torfw/tor-rules.new; exit 0)>/var/cache/torfw/tor-rules.diff

	while read line; do
	    origin="${line:0:1}"
	    val="${line:2}"
	    case $origin in
		\<)
		# in original, no longer in current
		    action=`echo "$val" | sed -e 's/ -A / -D /'`
		    eval "$action"
		    ;;
		\>)
		    action="$val"
		    eval "$action"
		    ;;
	    esac
	done < /var/cache/torfw/tor-rules.diff
    else
	$IPTABLES -t filter -F is_tor
	if [ -s /var/cache/torfw/tor-rules.new ]; then
	    . /var/cache/torfw/tor-rules.new
	else
	    # hmmm ... bootstrapping?
	    # use old if available
	    if [ -f /var/cache/torfw/tor-rules ]; then
		. /var/cache/torfw/tor-rules
	    fi
	fi
    fi

    if [ -s /var/cache/torfw/tor-rules.new ]; then
	mv /var/cache/torfw/tor-rules.new /var/cache/torfw/tor-rules
    fi
    
}

create_tor_rules()
{
    $IPTABLES -N is_tor
    add_tor_rules 'A'
}

update_tor_rules()
{
    add_tor_rules 'U'
}

add_trusted_interfaces()
{
    if [ -n "$trusted" ]; then
	for i in `echo "$trusted" | sed -e 's/,/\n/g'`; do
	    for c in INPUT OUTPUT; do
		$IPTABLES -t filter -A $c -i $i
	    done
	done
    fi
}

generate_rules() {

    parse_config_files

    set_default_policy
    flush_all_rules
    create_reject_drop_rules
    create_local_port_rules
    create_tor_rules
    add_trusted_interfaces

    $IPTABLES -A INPUT -m state --state INVALID -j drop_and_log
    $IPTABLES -A INPUT -m state --state RELATED,ESTABLISHED -j ACCEPT
    $IPTABLES -A INPUT -j local_ports 
    $IPTABLES -A INPUT -p icmp -s localhost -j ACCEPT
    $IPTABLES -A INPUT -j reject_and_log

    $IPTABLES -A OUTPUT -m state --state RELATED,ESTABLISHED -j ACCEPT
    $IPTABLES -A OUTPUT -j local_ports 
    $IPTABLES -A OUTPUT -p icmp -d localhost -j ACCEPT
    $IPTABLES -A OUTPUT -j is_tor
    $IPTABLES -A OUTPUT -j reject_and_log
}

if [ 0 != "$EUID" ]; then
    echo "must be root to run" 2>&1
    exit 1
fi

if [ 1 = $update ]; then
    update_tor_rules
else
    echo "generating rules"
    generate_rules
fi

# eof
