#!/usr/bin/perl -w
#
# Copyright (c) 2007, Gregory Fleischer (gfleischer@gmail.com)
# 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# 3. Neither the name of the author nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

use strict;
use IO::File;
use Getopt::Long;

eval {
    require IO::Socket::Socks;
};
if ($@) {
    require "IO_Socket_Socks.pm";
}


local($|) = 1;

my $referer = undef;
my $host = undef;
my $useragent = undef;
my $request = undef;
my $method = undef;
my $post_data = undef;
my $post_file = undef;
my $output = undef;
my $output_overwrite = undef;
my $url_encode = 0;
my $overwrite = 0;
my $socksdebug = 0;
my $debug = 0;
my $help = 0;
my $server_response = 0;


my @headers = ();

if (!GetOptions(
	 "help" => \$help,
	 "headers=s" => \@headers,
	 "referer|referrer=s" => \$referer,
	 "host:s" => \$host,
	 "U|useragent|user-agent=s" => \$useragent,
	 "debug" => \$debug,
	 "socksdebug" => \$socksdebug,
	 "request=s" => \$request,
	 "method=s" => \$method,
	 "S|server-response" => \$server_response, # like wget
	 "post-data=s" => \$post_data,
	 "post-file=s" => \$post_file,
	 "O|output=s" => \$output,
	 "OO|output-overwrite=s" => \$output_overwrite,
	 "url-encode|urlencode" => \$url_encode,
    )) {
    &help();
    exit(1);
}
if ($help) {
    &help();
    exit(0);
}

my @args = @ARGV;

my $host_addr = shift(@args) or die "Missing host addr";
my $host_port = 80;
my $resource = "/";
my $is_secure = 0;

if ($host_addr=~m!^http(s?)://([^/:]+)(?::(\d+))?(/?.*)!) {
    $is_secure = $1 ? 1 : 0;
    $host_addr = $2;
    $host_port = $3 || ($is_secure ? 443 : 80);
    $resource = $4 || "/";
} else {
    $resource = (shift(@args) || "/");
    if ($host_addr =~ /^(.+?):(\d+)$/) {
	$host_addr = $1;
	$host_port = $2;
    }
}

# default to host based on address unless one specified
if (!defined($host)) {
    $host = $host_addr;
    $host =~ s/\.\$[A-F0-9]+\.exit$//;
}

# post data ?
my $content_length = undef;
if (defined($post_file)) {
    $method = "POST" unless defined($method);
    die "bad request file: $request" unless -f $post_file;
    my $fh = new IO::File("< $post_file") or die "failed to read file: $post_file ($!)";
    my @p =<$fh>;
    close($fh);
    $post_data = join("", @p);
    $content_length = length($post_data);
} elsif (defined($post_data)) {
    $method = "POST" unless defined($method);
    $content_length = length($post_data);
}

# if method not set, assume GET
$method ||= "GET";

# output goes where ?
my $outfile = *STDOUT;
if (defined($output_overwrite)) {
    $output = $output_overwrite;
    $overwrite = 1;
}
if (defined($output)) {
    if ("-" eq $output) {
	$outfile = *STDOUT;
    } else {
	if (-f $output && !$overwrite) {
	    die "file [$output] exists";
	} else {
	    $outfile = new IO::File("> $output") or die "failed to open [$output] ($!)";
	}
    }
}

if ($url_encode) {
    my $qpos = index($resource, "?");
    if (-1 != $qpos) {
	my $path = substr($resource, 0, $qpos);
	my $query = substr($resource, $qpos);
	$path=~s/([^-a-zA-Z\.\/])/sprintf("%%%02X", ord($1))/ge;
	$resource = $path . $query;
    } else {
	$resource=~s/([^-a-zA-Z\.\/])/sprintf("%%%02X", ord($1))/ge;
    }
}

# create request
my @req = ();
if (defined($request)) {
    die "bad request file: $request" unless -f $request;
    my $fh = new IO::File("< $request") or die "failed to read file: $request ($!)";
    @req =<$fh>;
    close($fh);
    my %override = (
	"referer" => $referer,
	"useragent" => $useragent,
	"referrer" => $referer,
	"user-agent" => $useragent,
	"host" => $host,
	"resource" => $resource,
	"method" => $method,
	);
    my $interpolate = sub {
	my $n = shift();
	return $override{$n} if defined($override{$n});
	return $ENV{$n} if defined($ENV{$n});
	return $override{lc($n)} if defined($override{lc($n)});
	die "Bad substituion {$n}";
    };
    foreach (@req) {
	chomp;
	s/\$\{(\w+)\}/&{$interpolate}($1)/ge;
    }
} else {
    push(@req, "$method $resource HTTP/1.0");
    push(@req, "Host: $host");
    push(@req, "Connection: close");
    if (defined($referer)) {
	push(@req, "Referer: $referer");
    }
    if (defined($useragent)) {
	push(@req, "User-Agent: $useragent");
    }
    foreach my $header (@headers) {
	push(@req, $header);
    }
    if (defined($content_length)) {
	push(@req, "Content-Length: $content_length");
    }
}

# open socket via SOCKS
# XXX: should really read from torrc
my $proxy_addr = $ENV{SOCKS_HOST} || '127.0.0.1';
my $proxy_port = $ENV{SOCKS_PORT} || 9050;
my $socket = new IO::Socket::Socks(
    ProxyAddr => $proxy_addr,
    ProxyPort => $proxy_port,
    ConnectAddr => $host_addr,
    ConnectPort => $host_port,
    SocksDebug => $socksdebug
    ) or die "connect to [$host_addr:$host_port] failed";

# XXX: need SSL support

# send headers
foreach my $req (@req) {
    print STDERR $req, "\n" if $debug;
    print $socket $req, "\r\n";
}
print STDERR "\n" if $debug;
print $socket "\r\n";

# send post body
if (defined($post_data)) {
    print STDERR $post_data, "\n" if $debug;
    print $socket $post_data;
}

# read response
my $fc = 0;
while (<$socket>) {
    if ($fc) {
	print $outfile $_;
    } elsif ($server_response) {
	print STDERR $_;
    }
    if ("\r\n" eq $_) {
	$fc = 1;
    }
}

exit(0);

#
sub help {
    print "********************************************************************", "\n";
    print " socks-http.pl \n";
    print " Copyright (c) 2007, Gregory Fleischer (gfleischer\@gmail.com) \n";
    print "********************************************************************", "\n";
    print "\n";
    print "usage: $0 [options] URI\n";
    print "\n";
    print "   make HTTP request via Tor\n";
    print "\n";
    print "options:\n";
    print " -O, --output=<file>             Output (defaults to STDOUT)\n";
    print " -OO, --output-overwrite=<file>  Output and overwrite\n";
    print " --referer=<referer>\n";
    print " -U, --user-agent=<UA>           User-agent\n";
    print " --host=<host>\n";
    print " --method=<method>               GET,HEAD,TRACE,etc.\n";
    print " --debug                         print request\n";
    print " --socksdebug                    enable SOCKS debugging\n";
    print " --request=<file>                read request from file\n";
    print " -S, --server-response           print server response\n";
    print " --post-data=<data>              send post data\n";
    print " --post-file=<file>              send post data from file\n";
    print " --header=<header>               Added HTTP header (can be repeated)\n";
    print " --url-encode                    URL encode the request PATH\n";
    print " --help                          Display this help\n";
    print "\n";
}

# eof
